/*
 *  Copyright 2001-2003 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

%{

#include <iostream>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include "ragel.h"
#include "parsetree.h"

using std::endl;

/* The parse data. For each fsm spec, the parser collects things that it parses
 * in data structures in here. */
ParseData *pd;

/* Bison 1.875 tries to prevent a warning, but results ends up producing a 
 * This is an ugly hack. But it works for this case. */
#ifdef BISON_V1875
#define __attribute__(arg)
#endif

/* Try to do a definition, common to assignment and instantiation. */
void tryMachineDef( const BISON_YYLTYPE &loc, char *name, 
		ExpressionNode *exprNode, bool isInstance );

%}

%union {
	/* General data types. */
	char *data;
	int integer;
	Literal *literal;

	/* Tree nodes. */
	TermNode *term;
	FactorWithRepNode *factorWithRep;
	FactorWithAugNode *factorWithAug;
	FactorWithNegNode *factorWithNeg;
	FactorNode *factor;
	ExpressionNode *expression;

	/* Priorities and actions. */
	AugType augType;
	int actionId;
	PriorDesc *priorDesc;

	/* Regular expression items. */
	RegExp *regExp;
	ReItem *reItem;
	ReOrBlock *reOrBlock;
	ReOrItem *reOrItem;

	/* Explicit machine. */
	ExplicitMachine *explicitMach;
}

%token RL_Section

%token <data> RL_UInt
%token <data> RL_PInt
%token <data> RL_NInt
%token <data> RL_Hex
%token <data> RL_Word
%token <data> RL_Literal
%token RL_DoubleDot
%token RL_Arrow
%token RL_ColonEquals

%token RL_ReSlash
%token RL_ReSqOpen
%token RL_ReSqOpenNeg
%token RL_ReSqClose
%token RL_ReDot
%token RL_ReStar
%token RL_ReDash
%token <data> RL_ReChar

%token <data> RL_Builtin
%token RL_Data
%token RL_Action
%token RL_Init
%token RL_AlphType
%token RL_Range
%token RL_Exp
%token RL_Start
%token RL_Final
%token '*' '?' '+' '!' '(' ')' ';' '=' ':' '%' '$' '-' '|' '&' '.' '>'

%type <augType> AugType
%type <integer> PriorityAug
%type <data> PriorityAugStr
%type <actionId> ActionEmbed
%type <expression> Expression 
%type <term> Term
%type <factorWithAug> FactorWithAug 
%type <factorWithRep> FactorWithRep 
%type <factorWithNeg> FactorWithNeg
%type <factor> Factor
%type <literal> RangeLit
%type <data> AlphabetNum
%type <explicitMach> ExplicitMachine
%type <integer> ExpStateMods
%type <data> MachineName;
%type <integer> OptPriorName;

%type <regExp> RegularExpr
%type <reItem> RegularExprItem
%type <reItem> RegularExprChar
%type <reOrBlock> RegularExprOrData
%type <reOrItem> RegularExprOrChar

%%

/* Input is any number of input sections. An empty file is accepted. */
input: FsmSpecList ;
FsmSpecList: 		
	FsmSpecList FsmSpec { } |
	{ };

/* Fsm Specification. Fsms are deliminated by the %% tokens 
 * and require a name for the fsm immediately following the open %% 
 * deliminator. */
FsmSpec:
	RL_Section RL_Word StatementList RL_Section {
		// Set the fsm's name. */
		pd->fsmName = $2;
		pd->fsmStartSecLoc = InputLoc(@1);
		pd->fsmEndSecLoc = InputLoc(@4);

		// Generate output, dot file or fsm spec. */
		switch ( outputFormat ) {
		case OutCCode: case OutCppCode: 
			pd->generateCode( );
			break;
		case OutGraphvizDot:
			pd->generateGraphviz( );
			break;
		case OutPrint:
			pd->printFsm();
			break;
		case OutDump:
			pd->dumpFsm();
			break;
		}
		delete pd;

		// Make a parse data for the next fsm machine.
		pd = new ParseData( inputFile );
	};

/* Garble up a to the end of a section. */
FsmSpec: RL_Section error RL_Section { yyerrok; };


/* A NonEmpty list of statements in a fsm. */
StatementList:
	StatementList Statement |
	/* Nothing */;

/* The differnt types of statements in a fsm spec. */
Statement:	
	Assignment |
	Instantiation |
	DataSpec |
	ActionSpec |
	InitSpec |
	AlphSpec |
	RangeSpec;

/* Garble up to the next ; */
Statement: error ';' { yyerrok; };

/* Allow the user to give some code to insert into the init function. */
InitSpec:
	RL_Init RL_Builtin {
		// Put the code on the init code list. 
		pd->initCodeList.append( new StringListEl( InputLoc(@2), $2 ) );
	} ;

/* Allow the user to give some declarations to put into the fsm structure. */
DataSpec:	
	RL_Data RL_Builtin ';' {
		// Put the spec for the data on the list. 
		pd->dataList.append( new StringListEl( 
				InputLoc(@2), $2 ) );
	} ;

/* Allow the user to create a named fsm action that can be referenced when
 * building a machine. */
ActionSpec:	
	RL_Action RL_Word RL_Builtin {
		if ( pd->actionDict.insert( $2, pd->nextActionId ) ) {
			// Add the action to the list of actions.
			pd->actionList.append( new ActionListEl( InputLoc(@3), $2, $3, pd->nextActionId ) );

			// Used up an action id.
			pd->nextActionId += 1;
		}
		else {
			// Recover by just ignoring the duplicate.
			error(@2) << "action \"" << $2 << "\" already defined" << endl;
		}
	} ;

/* Specifies the data type of the input alphabet. One or two words 
 * followed by a semi-colon. */
AlphSpec:
	RL_AlphType RL_Word RL_Word ';' {
		if ( ! pd->setAlphType( $2, $3 ) ) {
			// Recover by ignoring the alphtype statement.
			error(@2) << "\"" << $2 << 
				" " << $3 << "\" is not an allowed alphabet type" << endl;
		}
	} |
	RL_AlphType RL_Word ';' {
		if ( ! pd->setAlphType( $2 ) ) {
			// Recover by ignoring the alphtype statement.
			error(@2) << "\"" << $2 << "\" is not an allowed alphabet type" << endl;
		}
	};

/* Specifies a range to assume that the input characters will fall into. */
RangeSpec:
	RL_Range AlphabetNum AlphabetNum ';' {
		// Save the upper and lower ends of the range and emit the line number.
		pd->lowerNum = $2;
		pd->upperNum = $3;
		pd->rangeLowLoc = InputLoc(@2);
		pd->rangeHighLoc = InputLoc(@3);
	};

/* An assignement statement. Assigns the definition of a machine to a variable name. */
Assignment:
	MachineName '=' Expression ';' {
		/* Generic creation of machine for instantiation and assignment. */
		tryMachineDef( @1, $1, $3, false );
	};

/* An instantiation statement. Instantiates a machine and assigns it to a variable name. */
Instantiation:
	MachineName RL_ColonEquals Expression ';' {
		/* Generic creation of machine for instantiation and assignment. */
		tryMachineDef( @1, $1, $3, true );
	};

/* Capture the machine name for making the machine's priority name. */
MachineName:
	RL_Word {
		// Make the machine available as a priority name. The name may be used already.
		PriorDictEl *priorDictEl;
		if ( pd->priorDict.insert( $1, pd->nextPriorKey, &priorDictEl ) )
			pd->nextPriorKey += 1;

		// Set the prior num inserted/found as the current default priority
		// number.
		pd->curDefPriorKey = priorDictEl->value;
	};

/* Top level production in the parse of a fsm. The lowest precedence
 * is the '|' (or), '&' (intersection), and '-' (subtraction) operators. */
Expression:
	Expression '|' Term {
		$$ = new ExpressionNode( $1, $3, ExpressionNode::Or );
	} |
	Expression '&' Term {
		$$ = new ExpressionNode( $1, $3, ExpressionNode::Intersect );
	} |
	Expression '-' Term {
		$$ = new ExpressionNode( $1, $3, ExpressionNode::Subtract );
	} |
	Term {
		$$ = new ExpressionNode( $1 );
	};

/* The second group up in precedence. The concatenation operators. */
Term:
	Term FactorWithAug {
		$$ = new TermNode( $1, $2 );
	} |
	Term '.' FactorWithAug {
		$$ = new TermNode( $1, $3 );
	} |
	FactorWithAug {
		$$ = new TermNode( $1 );
	};

/* Third group up in precenedce. Allow users to embed actions and to change the
 * priorities of transitions. */
FactorWithAug:
	FactorWithAug AugType ActionEmbed {
		$$ = $1;
		$$->actions.append( ParserAction( $3, 0, $2 ) );
	} |
	FactorWithAug AugType OptPriorName PriorityAug {
		// Append the priority augmentation to to the vector of priority augs
		// in the factor with aug and return the factor with aug.
		$1->priorityAugs.append( PriorityAug( $2, $3, $4 ) );
		$$ = $1;
	} |
	FactorWithRep {
		$$ = new FactorWithAugNode( $1 );
	};

/* Priority change specs. */
PriorityAug: 
	PriorityAugStr {
		// Convert the priority number to a long. Check for overflow.
		errno = 0;
		int aug = strtol( $1, 0, 10 );
		if ( errno == ERANGE && aug == LONG_MAX ) {
			// Priority number too large. Recover by setting the priority to 0.
			error(@1) << "priority number " << $1 << " overflows" << endl;
			$$ = 0;
		}
		else if ( errno == ERANGE && aug == LONG_MIN ) {
			// Priority number too large in the neg. Recover by using 0.
			error(@1) << "priority number " << $1 << " underflows" << endl;
			$$ = 0;
		}
		else {
			// No overflow or underflow.
			$$ = aug;
 		}
	};

/* Optional priority name. */
OptPriorName:
	RL_Word ',' { 
		// Lookup/create the priority key.
		PriorDictEl *priorDictEl;
		if ( pd->priorDict.insert( $1, pd->nextPriorKey, &priorDictEl ) )
			pd->nextPriorKey += 1;

		// Use the inserted/found priority key.
		$$ = priorDictEl->value;
	} |
	/* Nothing */ { 
		// No priority name, use the default.
		$$ = pd->curDefPriorKey; 
	};

PriorityAugStr:
	RL_NInt | RL_PInt | RL_UInt ;


		
/* Differnt places to change the priorities or to embed actions.  Used by
 * FactorWithAug for both types of augmentation. */
AugType:
	':' { $$ = fin; } |
	'%' { $$ = leave; } |
	'$' { $$ = all; } |
	'>' { $$ = start; };

/* Different ways to embed actions. A RL_Word is reference to an action given by
 * the user as a statement in the fsm specification. An action can also be
 * specified immediately. */
ActionEmbed:
	RL_Word {
		// Set the name in the actionDict.
		ActionDictEl *fdel = pd->actionDict.find( $1 );
		if ( fdel != 0 ) {
			// Pass up the action id;
			$$ = fdel->value;
		}
		else {
			// Recover by returning null as the action.
			error(@1) << "action lookup of \"" << $1 << "\" failed" << endl;
			$$ = 0;
		}
	} |
	RL_Builtin {
		// Add the action to the list of actions.
		pd->actionList.append( new ActionListEl( InputLoc(@1), 0, $1, pd->nextActionId ) );

		// Pass up the builtin name.
		$$ = pd->nextActionId;
		pd->nextActionId += 1;
	};

/* The fourth level of precedence. These are the trailing unary operators that
 * allow for repetition. */
FactorWithRep:
	FactorWithRep '*' {
		$$ = new FactorWithRepNode( InputLoc(@2), $1, 
				FactorWithRepNode::Star );
	} |
	FactorWithRep '?' {
		$$ = new FactorWithRepNode( InputLoc(@2), $1, 
				FactorWithRepNode::Optional );
	} |
	FactorWithRep '+' {
		$$ = new FactorWithRepNode( InputLoc(@2), $1, 
				FactorWithRepNode::Plus );
	} |
	FactorWithNeg {
		$$ = new FactorWithRepNode( InputLoc(@1), $1 );
	};

/* The fifth level up in precedence. Negation. */
FactorWithNeg:
	'!' FactorWithNeg {
		$$ = new FactorWithNegNode( InputLoc(@1), $2 );
	} |
	Factor {
		$$ = new FactorWithNegNode( InputLoc(@1), $1 );
	};

/* The highest level in precedence. Atomic machines such as references to other
 * machines, literal machines, regular expressions or Expressions in side of
 * parenthesis. */
Factor:
	RL_Literal {
		// Create a new factor node going to a concat literal. */
		$$ = new FactorNode( new Literal( InputLoc(@1), $1, Literal::LitString ) );
	} |
	AlphabetNum {
		// Create a new factor node going to a literal number. */
		$$ = new FactorNode( new Literal( InputLoc(@1), $1, Literal::Number ) );
	} |
	RL_Word {
		// Find the named graph.
		GraphDictEl *gdNode = pd->graphDict.find( $1 );
		if ( gdNode == 0 ) {
			// Recover by returning null as the factor node.
			error(@1) << "graph lookup of \"" << $1 << "\" failed" << endl;
			$$ = 0;
		}
		else if ( gdNode->isInstance ) {
			// Recover by retuning null as the factor node.
			error(@1) << "references to graph instantiations not allowed "
					"in expressions" << endl;
			$$ = 0;
		}
		else {
			// Create a factor node that is a lookup of an expression.
			$$ = new FactorNode( gdNode->value, FactorNode::LookupExpression );
		}
	} |
	RL_ReSqOpen RegularExprOrData RL_ReSqClose {
		// Create a new factor node going to an OR expression. */
		$$ = new FactorNode( new ReItem( InputLoc(@1), $2, ReItem::OrBlock ) );
	} |
	RL_ReSqOpenNeg RegularExprOrData RL_ReSqClose {
		// Create a new factor node going to a negated OR expression. */
		$$ = new FactorNode( new ReItem( InputLoc(@1), $2, ReItem::NegOrBlock ) );
	} |
	RL_ReSlash RegularExpr RL_ReSlash {
		// Create a new factor node going to a regular exp.
		$$ = new FactorNode( $2 );
	} |
	RangeLit RL_DoubleDot RangeLit {
		// Create a new factor node going to a range. */
		$$ = new FactorNode( new Range( $1, $3 ) );
	} |
	ExplicitMachine {
		// Make a factor node witht the explicit machine.
		$$ = new FactorNode( $1 );
	} |
	'(' Expression ')' {
		$$ = new FactorNode( $2, FactorNode::Expression );
	};

/* Garble up to the closing brace of a parenthesized expression. */
Factor: '(' error ')' { $$ = 0; yyerrok; };

/* Any form of a number that can be used as a basic machine. */
AlphabetNum:
	RL_UInt | RL_PInt | RL_NInt | RL_Hex ;

/* Parser for regular expression fsms. Any number of expression items which
 * generally gives a machine one character long or one character long stared. */
RegularExpr:
	RegularExpr RegularExprItem {
		// An optimization to lessen the tree size. If a non-starred char is directly
		// under the left side on the right and the right side is another non-starred
		// char then paste them together and return the left side. Otherwise
		// just put the two under a new reg exp node.
		if ( $2->type == ReItem::Data && !$2->star &&
			$1->type == RegExp::RecurseItem &&
			$1->item->type == ReItem::Data && !$1->item->star )
		{
			// Append the right side to the right side of the left and toss 
			// the right side.
			$1->item->str += $2->str;
			delete $2;
			$$ = $1;
		}
		else {
			$$ = new RegExp( $1, $2 );
		}
	} |
	/* Nothing */ {
		// Can't optimize the tree.
		$$ = new RegExp();
	};

/* RegularExprItems can be a character spec with an optional staring of the char. */
RegularExprItem:
	RegularExprChar RL_ReStar {
		$1->star = true;
		$$ = $1;
	} |
	RegularExprChar {
		$$ = $1;
	};

/* A character spec can be a set of characters inside of square parenthesis,
 * a dot specifying any character or some explicitly stated character. */
RegularExprChar:
	RL_ReSqOpen RegularExprOrData RL_ReSqClose {
		$$ = new ReItem( InputLoc(@1), $2, ReItem::OrBlock );
	} |
	RL_ReSqOpenNeg RegularExprOrData RL_ReSqClose {
		$$ = new ReItem( InputLoc(@1), $2, ReItem::NegOrBlock );
	} |
	RL_ReDot {
		$$ = new ReItem( InputLoc(@1), ReItem::Dot );
	} |
	RL_ReChar {
		$$ = new ReItem( InputLoc(@1), $1[0] );
	};

/* The data inside of a [] expression in a regular expression. Accepts any
 * number of characters or ranges. */
RegularExprOrData:
	RegularExprOrData RegularExprOrChar {
		// An optimization to lessen the tree size. If an or char is directly
		// under the left side on the right and the right side is another or
		// char then paste them together and return the left side. Otherwise
		// just put the two under a new or data node.
		if ( $2->type == ReOrItem::Data &&
				$1->type == ReOrBlock::RecurseItem &&
				$1->item->type == ReOrItem::Data )
		{
			// Append the right side to right side of the left and toss
			// the right side.
			$1->item->str += $2->str;
			delete $2;
			$$ = $1;
		}
		else {
			// Can't optimize, put the left and right under a new node.
			$$ = new ReOrBlock( $1, $2 );
		}
	} | 
	/* Nothing */ {
		$$ = new ReOrBlock();
	};


/* A single character inside of an or expression. Can either be a character
 * or a set of characters. */
RegularExprOrChar:
	RL_ReChar {
		$$ = new ReOrItem( $1[0] );
	} |
	RL_ReChar RL_ReDash RL_ReChar {
		$$ = new ReOrItem( $1[0], $3[0] );
	};

RangeLit:
	RL_Literal {
		// Range literas must have only one char.
		if ( strlen($1) != 1 ) {
			// Recover by using the literal anyways.
			error(@1) << "literal used in range must be of length 1" << endl;
		}
		$$ = new Literal( InputLoc(@1), $1, Literal::LitString );
	} |
	AlphabetNum {
		// Create a new literal number.
		$$ = new Literal( InputLoc(@1), $1, Literal::Number );
	};

/* Explicit machine is the exp keyword then a list of states in paren. */
ExplicitMachine:
	RL_Exp '(' ExpStateList ')' {
		// This will transfer the state list and copy the actions.
		$$ = new ExplicitMachine( InputLoc(@1), 
				pd->expStateList );
	};

/* A list of comma separated states. Requires at least one state. */
ExpStateList: 
	ExpStateList ';' ExpState |
	ExpState;

/* Explict state machine state. Makes a new state with the collected
 * transitions and appends it to the state list. */
ExpState:
	ExpStateMods RL_Word ExpStateAug ExpTransList {
		// Make the new state and append it.
		pd->expStateList.append( new ExpState( 
				InputLoc(@2),
				$1&ESM_START, 
				$1&ESM_FINAL,
				$2, 
				pd->expOutActions, 
				pd->expIsOutPriorSet,
				pd->expOutPriority,
				pd->expTransList ));

		// Clear out the collected data.
		pd->expOutActions.empty();
		pd->expOutPriority = 0;
		pd->expIsOutPriorSet = false;
		pd->expTransList.empty();
	};

/* Possible state modifications. A state can be start, final, both or neither. */
ExpStateMods:
	RL_Start RL_Final { $$ = ESM_START | ESM_FINAL; } |
	RL_Final RL_Start { $$ = ESM_START | ESM_FINAL; } |
	RL_Start { $$ = ESM_START; } |
	RL_Final { $$ = ESM_FINAL; } |
	/* Nothing */ { $$ = 0; };


/* Augmentation at the end of a state. Can specify out actions or the out
 * priority of the state */
ExpStateAug:
	ExpStateAug '%' ActionEmbed {
		// Save the action id.
		pd->expOutActions.append( $3 );
	} |
	ExpStateAug '%' PriorityAug {
		// Save the priority aug for the state.
		pd->expOutPriority = $3;
		pd->expIsOutPriorSet = true;
	} |
	/* Nothing */;

/* Optional transition list. */
ExpTransList:
	':' ExpTLOneOrMore {
	} |
	/* Nothing. */;

/* A transition list with one or more transitions. Needed to properly get
 * commas. */
ExpTLOneOrMore:
	ExpTLOneOrMore ',' ExpTrans |
	ExpTrans;

/* A transition. Supports single, range and default transitions. */
ExpTrans:
	Expression RL_Arrow RL_Word {
		// Append the transition to the vector of single transitions.
		pd->expTransList.append( ExpTrans( $1, $3 ) );
	};

%%

/* Try to do a definition, common to assignment and instantiation. Warns about 
 * instances other than main not being implemented yet. */
void tryMachineDef( const BISON_YYLTYPE &loc, char *name, ExpressionNode *exprNode, 
		bool isInstance )
{
	GraphDictEl *newEl = pd->graphDict.insert( name );
	if ( newEl != 0 ) {
		/* New element in the dict, all good. */
		newEl->value = exprNode;
		newEl->isInstance = isInstance;
		newEl->line = loc.first_line;
		newEl->col = loc.first_column;

		/* Issue a warning for instances other than main. */
		if ( isInstance && strcmp(name, machineMain) != 0 )
			warning(loc) << "instances other than \"main\" not yet implemented" << endl;
	}
	else {
		// Recover by ignoring the duplicate.
		error(loc) << "variable \"" << name << "\" previously defined" << endl;
	}
	pd->machineGiven = true;
}

void yyerror( char *err )
{
	error(yylloc.first_line, yylloc.first_column) << err << endl;
}
